{****************************************************************************
 * SoundLib 2 Soundblaster driver - (c) 1997-2001 by Stefan Goehler         *
 * More or less only simple rewrite of SoundLib 1.x, but does its work.     *
 * Improvements will come soon - didn't have time yet for them.             *
 * Copyright agreement in soundlib.txt - illegal copy without this file!    *
 * Changes over old version:                                                *
 * * Better playback on SB Live!                                            *
 * + automatic setting of playback options                                  *
 ****************************************************************************}
{$IFDEF FPC}
{$A+,B-,F+,G+,N+,E-,P+,Q-,R-,T-,V+,X+}
{$ASMMODE INTEL}
{$ELSE}
{$A+,B-,F+,G+,N+,E-,P+,Q-,R-,T-,V+,X+}
{$ENDIF}

{ History
  a2:
  .removed SB Pro 2.0 support due to problems with SB Pro emulation cards
  .automatic SB 16 settings detection (used if no environment variable)
  .soundblaster volume controls available

  a3
  .made SB16 mode more compatible with clone cards
  .FPU/MMX commands aren't disturbed anymore, registers are being saved
  .during interrupts
  .does SB16 initialization automatically if card hasn't been initialized
   yet
   -only on non-PnP SB16+
   -sets to IRQ 5, LoDMA 1, HiDMA 5
  .unit ports used for FPC compatible code
  .made SB reset wait longer - sometimes 'no soundcard found' appeared


}

unit drv_sb;

interface
uses sndlib2;

type
   psoundblaster = ^tsoundblaster;
   tsoundblaster = object(tsounddriver)
     constructor init(bufsize : word);
     destructor  done;virtual;

     function    getcapabilities(var tcap : tcardinfo) : boolean;virtual;
     function    startplayback(freq,flags : longint) : boolean;virtual;
     procedure   stopplayback;virtual;
     function    getversionstring : string;virtual;
{
     function    startrecord(freq,flags : longint) : boolean;virtual;
     procedure   stoprecord;virtual;}

     function    getnumofoutcontrols : word;virtual;
     function    getoutcontroltype(num : word) : word;virtual;
     function    getoutcontrolname(cnt : word) : string;virtual;
     function    getoutcontrolrange(cnt : word) : longint;virtual;
     function    getoutcontrol(cnt : word) : longint;virtual;
     function    setoutcontrol(cnt : word;val : longint) : boolean;virtual;
     procedure   intrcallproc;virtual;
     private
       controlmapping : array[0..15] of word;

       settings1 : array[1..15] of record
         l,r : byte;
       end;
       settings2 : array[1..8] of byte;

       maxrange : word;
       maxout   : word;

     function    getnum(cnt : word) : word;
     procedure   savemixer;
     procedure   restoremixer;
   end;

implementation
{$IFNDEF FPC}
uses dos,winapi,newfrontier,snd_cnv,ports,objects;
{$ELSE}
uses dos,snd_cnv,go32,ports,objects;
{$ENDIF}

type
  tplay = record
    hispeed    : boolean;{ SbPro internal usage }

    blocksize  : longint;{ half size of the getted buffermemory            }
    p          : longint;{ DOS pointer to the sounddata buffermemory       }
    fpcp       : pointer;
    activeblock: byte;   { active bufferblock (because of doublebuffering) }
    callproc   : procedure;
    callt      : psoundblaster;

    flags      : longint;

    frequency  : longint;{ samplingrate                }
  end;

  trec = record
    hispeed    : boolean; { SbPro internal usage }

    blocksize   : longint;{ half size of the getted buffermemory       }
    p           : longint;{ pointer to the sounddata buffermemory      }
    activeblock : byte;   { active bufferblock (doublebuffering)       }
    callproc    : procedure;{the procedure to call with the interrupt  }

    frequency   : longint;{ samplingrate                               }
    flags       : longint;
  end;

const
  sblive             = $40D;
  sb16               = $400;
  sbpro              = $300;
  sb21               = $210;
  sb20               = $200;

  { Constants for the dma-port-adresses }
  dma_lpage          : array[0..7] of byte = ($87,$83,$81,$82,$88,$8B,$89,$8A);
  dma_hpage          : array[0..7] of word = ($487,$483,$481,$482,$00,$48B,$489,$48A);
  dma_adr            : array[0..7] of byte = ($00,$02,$04,$06,$C0,$C4,$C8,$CC);
  dma_wc             : array[0..7] of byte = ($01,$03,$05,$07,$C2,$C6,$CA,$CE);

  buffersize  : word = 1024;   { Default playback buffer size in samples/2 }
  autorestore : boolean = true;{ automatic restore mixersettings }
  mixsaved    : boolean = false;

var
  oldexitproc   : pointer;

  blaster       : record
    irq         : byte;
    lodma,hidma : word;
    address     : word;
    aweport     : word; { not really used by unit }
    available   : boolean;
    cap         : tcardinfo;
  end;

  play             : tplay;
  rec              : trec;

  {$IFNDEF FPC}
  oldint           : pointer;
  {$ELSE}
  oldint           : tseginfo;
  newint           : tseginfo;
  int_ds           : word; external name '___v2prt0_ds_alias';
  {$ENDIF}
  irqmsk,irqvec    : byte;

  fpustate         : pointer;

const
  versionstring = 'Sound Blaster driver v2.0a3';

  procedure exitrecord;forward;

{  function  setoutput(output : byte;left,right : boolean) : boolean;forward;
  function  setinput(input,channel : byte;left,right : boolean) : boolean;forward;
  function  getinput(input,channel : byte;var left,right : boolean) : boolean;forward;}

function  readmixer(reg: byte) : byte;forward;
procedure writemixer(reg, val: byte);forward;

function writedsp(v : byte) : boolean;assembler;
asm
  mov   dx,blaster.address
  add   dx,0Ch
  xor   cx,cx
  @lp:
    in    al,dx
    dec   cx
  jz    @err
  cmp   al,127
  ja    @lp
  jmp   @noerr
  @err:
     mov   al,0
     jmp   @end
  @noerr:
  mov   al,v
  out   dx,al
  mov   al,1
@end:
end;

var
  endtime : longint;

function currenttime : longint;assembler;
asm
  mov   ax,0
  int   1Ah
  mov   ax,dx
  mov   dx,cx
end;

function resetblaster : boolean;
var
  b : byte;

begin
  outp(blaster.address+$6,1);
  endtime := currenttime+3;
  while (endtime > currenttime) do;
  outp(blaster.address+$6,0);
  endtime := currenttime+3;
  repeat
    b := inp(blaster.address+$E);
    b := inp(blaster.address+$A);
  until (b = $AA)or(currenttime >= endtime);
  resetblaster := b = $AA;
end;

{procedure resetblaster_fast;
var
  b : byte;
begin
  port[blaster.address+$6] := 1;
  port[blaster.address+$6] := 0;
  b := port[blaster.address+$E];
  b := port[blaster.address+$A];
end;}


procedure setspeed(hz : byte);assembler;
asm
  push   40h
  call   writedsp
  mov    al,hz
  dec    al
  push   ax
  call   writedsp
end;

procedure setspeedSB16play(hz : word);assembler;
asm
  push   42h
  call   writedsp
  mov    ax,hz
  mov    al,ah
  xor    ah,ah
  push   ax
  call   writedsp
  mov    ax,hz
  xor    ah,ah
  push   ax
  call   writedsp
end;

procedure setspeedSB16record(hz : word);assembler;
asm
  push   41h
  call   writedsp
  mov    ax,hz
  mov    al,ah
  xor    ah,ah
  push   ax
  call   writedsp
  mov    ax,hz
  xor    ah,ah
  push   ax
  call   writedsp
end;

function setsamplingrate(hz : word) : word;
var
  mul : word;

begin
  setsamplingrate := hz;
  if (play.flags and snd_stereo <> 0) then mul := 2 else mul := 1;

  if (blaster.cap.version >= sb16) then setspeedSB16play(hz) else
  begin
    if (blaster.cap.version >= sbpro) then
      setspeed(256-(1000000 div (longint(hz)*mul)))
      else setspeed(256-(1000000 div (longint(hz))));
      setsamplingrate := round(hz*((1000000 div longint(hz))*(hz)/1000000));

    if (hz < 36000) then { Set lowpass filter for better sound quality }
    begin
      writemixer($0E,readmixer($0E) and not 32);
      if hz < 18000 then writemixer($0C,readmixer($0C) and not 8) {3,2 khz}
        else writemixer($0C,readmixer($0C) or 8); {8,8 khz-filter}
    end
    else writemixer($0E,readmixer($0E) or 32);

  end;
end;

procedure setsamplingraterec(hz : word);
var
  mul : word;

begin
  if (rec.flags and snd_stereo <> 0) then mul := 2 else mul := 1;
  if blaster.cap.version >= sb16 then setspeedSB16record(hz) else
  setspeed(256-(1000000 div (longint(hz)*mul)));
end;

procedure initdma(dma : byte;p : longint;size : word;get,autoinit : boolean);
var
  b         : byte;
  l         : longint;
  segm,offs : word;

begin
  p := p and $FFFF0000;
  l := p shr 12;
  segm := l shr 16;
  if (dma > 3) then
  offs := word(p shr 13)
  else
  offs := l and $FFFF;

  if get then b := $44 else b := $48;
  if autoinit then inc(b,$10);

  if (dma > 3) then
  begin
    outp($D4,dma);
    outp($D6,b+dma-4);
    outp($D8,0);
  end else
  begin
    outp($0A,dma+4);
    outp($0B,b+dma);
    outp($0C,0);
  end;
  outp(dma_adr[dma],(offs));
  outp(dma_adr[dma],(offs) shr 8);
  outp(dma_lpage[dma],segm);
  outp(dma_wc[dma],(size-1));
  outp(dma_wc[dma],(size-1) shr 8);
  if (dma > 3) then outp($D4,dma-4) else outp($0A,dma);
end;

procedure playsb(p : longint;dsize : word);
begin
  initdma(blaster.lodma,p,dsize,false,false);
  writedsp($14);
  writedsp((dsize-1));
  writedsp((dsize-1) shr 8);
end;

procedure initrecord(p : longint;buffersize,frequency : word;flags : longint);
var
  size,dma : word;

begin
  if not blaster.available then exit;
  rec.blocksize  := buffersize shr 1;
  rec.frequency  := frequency;
  rec.p          := p;
  rec.activeblock:= 0;{Block where the last data has been written to}
  rec.hispeed    := false;
  rec.flags      := flags or snd_rec;

  size := play.blocksize;

  setsamplingraterec(frequency);

  if (blaster.cap.version >= sb16) then
  begin
    size := rec.blocksize;
    if (rec.flags and snd_16bit <> 0) then
    begin
      size := size shr 1;
      dma := blaster.hidma;
    end else dma := blaster.lodma;

    initdma(dma,rec.p,size shl 1,true,true);
    if (rec.flags and snd_16bit <> 0) then
    begin
      writedsp($BE);
      if (rec.flags and snd_stereo <> 0) then writedsp($30) else writedsp($10);{signed!!!}
      end else begin
        writedsp($CE);
        if (rec.flags and snd_stereo <> 0) then writedsp($20) else writedsp($00);
      end;
      writedsp((size-1));
      writedsp((size-1) shr 8);
    end else if
  ((blaster.cap.version >= sb21)and((rec.frequency > 13000))or
  ((blaster.cap.version >= sbpro)and((rec.frequency > 23000)or(rec.flags and snd_stereo <> 0)))) then
  begin { sb21 or higher and sr > 13000 or sbpro and sr > 23000 }
    writedsp($D3);
    rec.hispeed := true;
    if (rec.flags and snd_stereo <> 0)and(blaster.cap.version >= sbpro) then{sbpro and stereo and sr >= 11025 and sr <= 22050}
    writedsp($A8);

    initdma(blaster.lodma,rec.p,size shl 1,true,true);
    setsamplingrate(rec.frequency);

    writedsp($48);
    writedsp((size-1));
    writedsp((size-1) shr 8);
    writedsp($98);
  end else
  begin { sb20 or higher, mono and sr <= 13000 }
    writedsp($D3);
    initdma(blaster.lodma,rec.p,size shl 1,true,true);
    setsamplingrate(rec.frequency);

    writedsp($48);
    writedsp((size-1));
    writedsp((size-1) shr 8);
    writedsp($2C);
  end;
end;

var
  dummycalled : boolean;

procedure dummycaller;far;
begin
  dummycalled := true;
end;


procedure exitblaster;far;
begin
  exitproc := oldexitproc;
  if blaster.available then
  begin
    if (rec.flags and snd_play <> 0) then play.callt^.stopplayback;
    if (rec.flags and snd_rec <> 0) then exitrecord;
    {$IFNDEF FPC}
    setintvec(irqvec+(blaster.irq and 7),oldint);
    {$ELSE}
    set_pm_interrupt(irqvec+(blaster.irq and 7),oldint);
    {$ENDIF}
    if (blaster.irq < 8) then
    outp($21,inp($21) or irqmsk) else outp($A1,inp($A1) or irqmsk);
    writedsp($D3);
    if (blaster.irq < 8) then outp($20,$20) else outp($A0,$20);
  end;
end;

procedure exitplay;
begin
  if (play.flags and snd_play <> 0) then exit;
  if (blaster.cap.version >= sb16) then
  begin
    if (play.flags and snd_16bit <> 0) then writedsp($D9) else writedsp($DA);
    resetblaster{_fast};
  end
  else if (play.hispeed) then
  begin
    resetblaster{_fast};
    writedsp($D3);
    writemixer($E,readmixer($E) and not 2);
  end else
  begin
    writedsp($D3);
    writedsp($DA);
    resetblaster{_fast};
  end;
  play.flags := play.flags and not snd_play;
  play.callproc := dummycaller;
end;

procedure exitrecord;
begin
  if (rec.flags and snd_rec <> 0) then
  begin
    if blaster.cap.version >= sb16 then
    begin
      if (rec.flags and snd_16bit <> 0) then writedsp($D9) else writedsp($DA);
      resetblaster;
    end
    else if rec.hispeed then
    begin
      resetblaster;
      writemixer($E,readmixer($E) and not 2);
    end else
    begin
      writedsp($DA);
      resetblaster;
    end;
    rec.flags  := rec.flags and not snd_rec;
    rec.callproc := dummycaller;
  end;
end;


procedure silence(var p; count : word);
 procedure fillword(var X; Count: Word; value : word);assembler;
 asm les di,x;mov cx,count;mov ax,value;rep stosw;end;
begin
  if (play.flags and snd_16bit <> 0)and(blaster.cap.flags and snd_16bit <> 0)
   then fillword(p,count shr 1,0) else fillword(p,count shr 1,$8080);
end;

var
  dsjump : word;
  downsample : procedure(p1,p2 : pointer;adder,size : word);

function outofspec(freq,flags : longint) : boolean;
begin
  outofspec := true;
  if (longint(freq) > blaster.cap.monoplay.max)and(blaster.cap.flags and snd_stereo = 0) then exit;
  if (flags and snd_stereo <> 0)and(blaster.cap.flags and snd_stereo = 0) then exit;
  if (flags and snd_16bit <> 0)and(blaster.cap.flags and snd_16bit = 0) then exit;
  if (flags and snd_stereo <> 0)and(freq > blaster.cap.stereoplay.max) then exit;
  if (flags and snd_stereo = 0)and(freq > blaster.cap.monoplay.max) then exit;
  outofspec := false;
end;

procedure soundplaycaller;far;
begin
  play.activeblock := (play.activeblock+1) and 1;
end;

procedure intcaller;interrupt;
var
  h   : byte;

begin
  {$IFNDEF FPC}
  asm
    { Disable interrupts }
    cli

    { Save ALL registers }
    db 66h;pushf
    db 66h;pusha

           push   es
           push   ds
    db 0Fh,0A8h {push gs}
    db 0Fh,0A0h {push fs}

    { Save FPU state     }
    cmp    test8087,0
    je     @nofpu
      les    di,fpustate
      fsave  es:[di]
    @nofpu:
  end;
  {$ELSE}
  asm
    { Disable interrupts }
    cli

    { Save ALL registers }
    pushfd
    push   ds
    push   ax
    mov    ax,cs:[int_ds]
    mov    ds,ax

    push   es
    push   gs
    push   fs

    mov    ax,[dosmemselector]
    mov    fs,ax
    pushad

    { Save FPU state     }
    cmp    test8087,0
    je     @nofpu
      mov    edi,fpustate
      fsave  ds:[edi]
    @nofpu:
  end;
  {$ENDIF}

  if (blaster.cap.version >= sb16) then
  begin
    h := readmixer($82);
    if (h and 1 <> 0) then h := inp(blaster.address+$E) else h := inp(blaster.address+$F);
    if (h and 3 <> 0) then
    begin
      if (rec.flags and snd_rec <> 0) then rec.callproc;
      if assigned(play.callt) then play.callt^.intrcallproc;
      if (play.flags and snd_play <> 0)and(assigned(play.callproc)) then play.callproc;
    end;
  end else
  begin
    h := inp(blaster.address+$E);
    if assigned(play.callt) then play.callt^.intrcallproc;
    if (play.flags and snd_play <> 0)and(assigned(play.callproc)) then play.callproc;
    if (rec.flags and snd_rec <> 0) then rec.callproc;
  end;
  if (blaster.irq > 7) then outp($A0,$20);
  outp($20,$20);
  play.activeblock := (play.activeblock+1) and 1;
  {$IFNDEF FPC}
  asm
    { Restore FPU state    }
    cmp    test8087,0
    je     @nofpu
      les    di,fpustate
      frstor es:[di]
    @nofpu:

    { Reload ALL registers }
    db 0Fh,0A1h {pop fs}
    db 0Fh,0A9h {pop gs}
           pop   ds
           pop   es

    db 66h;popa
    db 66h;popf
    { Enable interrupts    }
    sti
  end;
  {$ELSE}
  asm
    { Restore FPU state    }
    cmp    test8087,0
    je     @nofpu
      mov    edi,fpustate
      frstor ds:[edi]
    @nofpu:

    { Reload ALL registers }
    popad

    pop   fs
    pop   gs
    pop   es

    pop   ax
    pop   ds

    popfd
     { Enable interrupts    }
    sti
  end;
  {$ENDIF}
end;

function hex2word(s : string) : word;
const
  number : set of char=['0'..'9'];
var
  i    : word;
  w,w2 : word;
begin
  w := 0;
  for i := 1 to length(s) do
  begin
    if (s[i] in number) then
    w2 := ord(s[i])-48
    else
    w2 := ord(s[i])-55;
    inc(w,w2 shl ((length(s)-i)*4));
  end;
  hex2word := w;
end;

function getdspversion : word;
var
  b : byte;
  v : word;

begin
  writedsp($E1);
  repeat
    b := inp(blaster.address+$0E);
  until (b and $80 = 0);

  repeat
    b := inp(blaster.address+$0A);
  until (b <> $AA);
  v := word(b) shl 8;
  repeat
    b := inp(blaster.address+$0A);
  until (b <> $AA);
  getdspversion := v+b;
end;

function detectawe32(baseport : word) : boolean;
var
  scratch,tmp : word;

begin
  detectawe32 := false;
  outpw($0E02+baseport,224);
  tmp := inpw($0E00+baseport);
  scratch := inpw($0E00+baseport);
  if (scratch and $000F <> $0C) then exit;
  outpw($0E02+baseport,32+29);
  scratch := inpw($0A00+baseport);
  if (scratch and $007E <> $58) then exit;
  outpw($0E02+baseport,32+30);
  scratch := inpw($0A00+baseport);
  if (scratch and $0003 <> $03) then exit;
  detectawe32 := true;
end;

function detectblaster_norm : boolean;
var
  s,s2 : string;
  i    : word;
  i2   : integer;

begin
  detectblaster_norm := false;
  s := getenv('BLASTER');
  if s = '' then exit;              {Get blaster environment string...}
  for i := 1 to length(s)
  do s[i] := upcase(s[i]);

  i := pos('D',s);
  if i <> 0 then
  begin
    val(s[i+1],i,i2);
    blaster.lodma := i;
  end else exit;
  i := pos('H',s);
  if i <> 0 then
  begin
    val(s[i+1],i,i2);
    blaster.hidma := i;
  end else blaster.cap.version := sbpro;

  i := pos('I',s);
  if i <> 0 then
  begin
    val(s[i+1],i,i2);
    blaster.irq := i;
    i := pos('I',s);
    if (i+2 <= length(s))and(s[i+2] <> ' ') then
    begin
      blaster.irq := blaster.irq*10;
      val(s[i+2],i,i2);
      inc(blaster.irq,i);
    end;
  end else exit;

  i := pos('A',s);
  if i <> 0 then
  begin
    for i2 := 1 to 3 do s2[i2] := s[i+i2];
    s2[0] := #3;
    s2[4] := #0;
    blaster.address := hex2word(s2);
  end else exit;

  detectblaster_norm := true;
end;

function detectblaster_auto16 : boolean;
var
  i,i2  : word;

begin
  detectblaster_auto16 := false;

  { Try to locate base address }
  blaster.address := $180;
  repeat
    inc(blaster.address,$20);
  until (getdspversion <> $FFFF);

  { If not found or not SB16, exit }
  if (getdspversion < sb16)or(blaster.address = $280) then exit;

  { Read IRQ and DMA settings }
  i  := readmixer($80) and $0F;
  i2 := readmixer($81);

  { If settings not set, then try to do manual init }
  if (i = 0)and(i2 = 0) then
  begin
    { Set to IRQ 5, DMA 1 and 5 }
    writemixer($80,$02);
    writemixer($81,$22);

    { Read new settings }
    i  := readmixer($80) and $0F;
    i2 := readmixer($81);
  end;

  case (i and $F) of
    1 : blaster.irq := 2;
    2 : blaster.irq := 5;
    4 : blaster.irq := 7;
    8 : blaster.irq := 10;
    else exit;
  end;

  case (i2 and $B) of
    1 : blaster.lodma := 0;
    2 : blaster.lodma := 1;
    8 : blaster.lodma := 3;
    else exit;
  end;

  case ((i2 shr 4) and $E) of
    2 : blaster.hidma := 5;
    4 : blaster.hidma := 6;
    8 : blaster.hidma := 7;
    else exit;
  end;

  detectblaster_auto16 := true;
end;

procedure writemixer(reg,val : byte);assembler;
asm
  mov   dx,blaster.address
  add   dx,04h
  mov   al,reg
  out   dx,al
  inc   dx
  mov   al,val
  out   dx,al
end;

function readmixer(reg : byte) : byte;assembler;
asm
  mov   dx,blaster.address
  add   dx,04h
  mov   al,reg
  out   dx,al
  inc   dx
  in    al,dx
end;

function setoutput(output : byte;left,right : boolean) : boolean;
procedure setbit(bit : byte;on : boolean);
begin
  if on then
  writemixer($3C,readmixer($3C) or 1 shl bit) else
  writemixer($3C,readmixer($3C) and not (1 shl bit));
end;

begin
  setoutput := false;
  if blaster.cap.version < sb16 then exit;
  case output of
    cnt_mic  : setbit(0,left);
    cnt_cd   : begin;setbit(1,right);setbit(2,left);end;
    cnt_line : begin;setbit(3,right);setbit(4,left);end;
    else exit;
  end;
  setoutput := true;
end;

function setinput(input,channel : byte;left,right : boolean) : boolean;
{How to use this function?
 -channel is 1 for left, 2 for right
 -you can switch the input for left _and_ right for one channel
 -if you record mono, the left configuration will be used -
  so if you wanna record both left and right in mono, switch
  channel 1(left) to left and right
 -if the blaster is sbpro - you can't set anything off -
  only one thing on with the call you will set the called input on}
procedure setbit(reg,bit : byte;left,right : boolean);
begin
  if (bit = 0) then right := left;
  if right then
  writemixer(reg,readmixer(reg) or 1 shl bit) else
  writemixer(reg,readmixer(reg) and not (1 shl bit));

  if (bit <> 0) then
  if left then
  writemixer(reg,readmixer(reg) or 1 shl succ(bit)) else
  writemixer(reg,readmixer(reg) and not (1 shl succ(bit)));
end;

var
  b : byte;
begin
  setinput := false;
  if blaster.cap.version >= sb16 then
  begin
    b := channel+$3C;
    case input of
      cnt_mic  : setbit(b,0,left,false);
      cnt_cd   : setbit(b,1,left,right);
      cnt_line : setbit(b,3,left,right);
      cnt_midi : setbit(b,5,left,right);
      else exit;
    end;
  end else if (blaster.cap.version >= sbpro)and((left)or(right)) then
  begin
    case input of
      cnt_cd   : b := 1;
      cnt_mic  : b := 2;
      cnt_line : b := 3;
      else exit;
    end;
    writemixer($0C,readmixer($0C) or (b shl 1));
  end else exit;
  setinput := true;
end;

function getinput(input,channel : byte;var left,right : boolean) : boolean;
procedure getbit(reg,bit : byte;var left,right : boolean);
begin
  byte(right) := (readmixer(reg) and (1 shl bit)) shr bit;
  if bit > 0 then
  byte(left) := (readmixer(reg) and (1 shl succ(bit))) shr succ(bit);
  if bit = 0 then left := right;
end;

var
  b,b2 : byte;

begin
  getinput := false;
  if (blaster.cap.version >= sb16) then
  begin
    b := channel+$3C;
    case input of
      cnt_mic  : getbit(b,0,left,right);
      cnt_cd   : getbit(b,1,left,right);
      cnt_line : getbit(b,3,left,right);
      cnt_midi : getbit(b,5,left,right);
      else exit;
    end;
  end else if (blaster.cap.version >= sbpro) then
  begin
    b2 := readmixer($0C) shr 1 and 3;
    case input of
      cnt_cd   : b := 1;
      cnt_mic  : b := 2;
      cnt_line : b := 3;
      else exit;
    end;
    left := b2 = b;
  end else exit;
  getinput := true;
end;


function initplay(p : longint;buffersize,frequency : word;flags : longint) : word;
var
  size,dma     : word;
  savecallproc : procedure;

begin
  if not blaster.available then exit;
  play.flags      := (flags and not snd_command) or snd_play;

  play.blocksize  := buffersize shr 1;
  play.frequency  := frequency;
  play.p          := p;
  play.activeblock:= 0;
  play.hispeed    := false;

  size := play.blocksize;
  if (blaster.cap.version >= sb16) then
  begin
    initplay := setsamplingrate(play.frequency);

    if (play.flags and snd_16bit <> 0) then
    begin
      size := size shr 1;
      dma := blaster.hidma;
    end
    else dma := blaster.lodma;

    initdma(dma,play.p,size shl 1,false,true);
    if (play.flags and snd_16bit <> 0) then writedsp($B6)
      else writedsp($C6);

    if (flags and snd_signed <> 0) then
    begin
      if (play.flags and snd_stereo <> 0) then writedsp($30) else writedsp($10) end
      else begin if (play.flags and snd_stereo <> 0) then writedsp($20) else writedsp($00);end;
      writedsp((size-1));
      writedsp((size-1) shr 8);
    end
    else if (blaster.cap.version >= sb21)and((play.frequency > 23000)
    or(play.flags and snd_stereo <> 0)) then
    begin                                      {sb21 or higher and sr > 23000}
    writedsp($D1);
    play.hispeed := true;
    if (play.flags and snd_stereo <> 0)and(blaster.cap.version >= sbpro) then
    begin                   {sbpro and stereo and sr >= 11025 and sr <= 22050}
      {This part switches the sbpro to stereo mode
       This should work according to the manual from
       creative labs}
      writemixer($E,readmixer($E) or 2);
      savecallproc  := play.callproc;
      dummycalled   := false;
      play.callproc := dummycaller;
      mem[play.p and $FFFF:0] := $80;
      playsb(play.p,1);
      repeat until dummycalled;
      play.callproc := savecallproc;
    end;

    initdma(blaster.lodma,play.p,size shl 1,false,true);
    initplay := setsamplingrate(play.frequency);

    writedsp($48);
    writedsp((size-1));
    writedsp((size-1) shr 8);
    writedsp($90);
  end else
  begin                                  {sb20 or higher, mono and sr <=23000}
    writedsp($D1);
    initdma(blaster.lodma,play.p,size shl 1,false,true);
    initplay := setsamplingrate(play.frequency);
    if (play.flags and snd_stereo <> 0) then writemixer($0E,readmixer($0E) or $20);
    writedsp($48);
    writedsp((size-1));
    writedsp((size-1) shr 8);
    writedsp($1C);
  end;
end;

constructor tsoundblaster.init(bufsize : word);
var
  i      : integer;
  p,oldp : longint;
  fpcp   : pointer;

begin
  tobject.init;
  if (bufsize >= 8192) then bufsize := 8191;
  if (bufsize = 0) then bufsize := buffersize;
  data.bufsize := bufsize;

  if blaster.available then fail;

  blaster.cap.version := 0;

  if (not detectblaster_norm) then
    if (not detectblaster_auto16) then fail;

  if not resetblaster then       { if blaster could not be resetted, }
  fail;                          { it must be busy                   }
  i := getdspversion;
  if blaster.cap.version = 0 then blaster.cap.version := i else
  if blaster.cap.version < sb16 then blaster.cap.version := i;
  {for any tests with lower blaster versions, change here the version}
{  blaster.cap.version := sb16;}
  if blaster.cap.version < sb20 then      { blasters lower than version  }
    fail;                                 { 2.0 aren't supported because }
                                          { they cannot run the autoinit }
                                          { mode - quality is too low    }

  { Set all required hardware capabilities }
  with blaster.cap do
  begin
    if (version >= sblive) then
    begin
      minfreq         := 4000;
      monoplay.pref   := 44100;
      monoplay.max    := 48000;
      stereoplay.pref := 44100;
      stereoplay.max  := 48000;
      monorecfreq     := 48000;
      stereorecfreq   := 48000;
      flags           := snd_16bit or snd_signed or snd_stereo;
      blaster.aweport := 0;
      maxrange        := 31;
      maxout          := 11;
      cardstring      := 'Soundblaster Live!';
    end else
    if (version >= sb16) then
    begin
      minfreq         := 5000;
      monoplay.pref   := 44100;
      monoplay.max    := 45000;
      stereoplay.pref := 44100;
      stereoplay.max  := 45000;
      monorecfreq     := 44100;
      stereorecfreq   := 44100;
      flags           := snd_16bit or snd_signed or snd_stereo;
      blaster.aweport := 0;
      maxrange        := 31;
      maxout          := 11;
      for i := 0 to 4 do if
      detectawe32(i*$20) then blaster.aweport := $600+i*$20;
      if (blaster.aweport = 0) then
      cardstring := 'Soundblaster 16' else
      cardstring := 'Soundblaster AWE32/64'
    end else
    if (version >= sbpro) then
    begin
      minfreq         := 4000;
      monoplay.pref   := 44100;
      monoplay.max    := 44100;
      stereoplay.pref := 23000;
      stereoplay.max  := 23000;
      monorecfreq     := 44100;
      stereorecfreq   := 22050;
      maxrange        := 15;
      maxout          := 5;
      flags           := snd_stereo;
      cardstring      := 'Soundblaster Pro'
    end else
    if (version >= sb21) then
    begin
      minfreq         := 4000;
      monoplay.pref   := 44100;
      monoplay.max    := 44100;
      stereoplay.pref := 0;
      stereoplay.max  := 0;
      monorecfreq     := 15000;
      stereorecfreq   := 0;
      maxrange        := 7;
      maxout          := 4;
      flags           := snd_default;
      cardstring      := 'Soundblaster 2.1'
    end else
    begin
      minfreq         := 4000;
      monoplay.pref   := 23000;
      monoplay.max    := 23000;
      stereoplay.pref := 0;
      stereoplay.max  := 0;
      monorecfreq     := 13000;
      stereorecfreq   := 0;
      maxrange        := 0;
      maxout          := 0;
      flags           := snd_default;
      cardstring      := 'Soundblaster 2.0'
    end;
  end;

  for i := 0 to maxout-1 do controlmapping[i] := i;

  blaster.cap.driverstring := versionstring;

  getmem(fpustate,4096);


  irqmsk := 1 shl (blaster.irq and 7);

  if (blaster.irq < 8) then
  irqvec := $8 else irqvec := $70;
  asm cli end;
  if (blaster.irq < 8) then
  outp($21,inp($21) or irqmsk) else
  outp($A1,inp($A1) or irqmsk);

  {$IFNDEF FPC}
  getintvec(irqvec+(blaster.irq and 7),oldint);
  setintvec(irqvec+(blaster.irq and 7),addr(intcaller));
  {$ELSE}
  newint.offset  := addr(intcaller);
  newint.segment := get_cs;
  get_pm_interrupt(irqvec+(blaster.irq and 7),oldint);
  set_pm_interrupt(irqvec+(blaster.irq and 7),newint);
  {$ENDIF}
  if (blaster.irq < 8) then
  outp($21,inp($21) and not irqmsk) else
  outp($A1,inp($A1) and not irqmsk);
  asm sti end;
  blaster.available := true;

  { Test if SB works }
  dummycalled   := false;
  play.callproc := dummycaller;

  {$IFNDEF FPC}
  p := globaldosalloc($100);
  fillchar(ptr(p and $FFFF,0)^,$100,$7F);
  {$ELSE}
  p := global_dos_alloc($100);
  getmem(fpcp,$100);
  fillchar(fpcp^,$100,$7F);
  dpmi_dosmemput(p shr 16,0,fpcp^,$100);
  {$ENDIF}
  oldp := play.p;

  initplay(p,$100,22050,snd_8bit+snd_mono+snd_nonsigned);
  endtime := currenttime+5;
  while (not dummycalled)and(currenttime < endtime) do;
  if not dummycalled then
  begin
    blaster.available := false;
    {$IFNDEF FPC}
    globaldosfree(p);
    {$ELSE}
    freemem(fpcp);
    global_dos_free(p);
    {$ENDIF}
    play.p := oldp;
    freemem(fpustate,4096);
    fail;
  end;
  resetblaster;
  {$IFNDEF FPC}
  globaldosfree(p);
  {$ELSE}
  global_dos_free(p);
  {$ENDIF}
  exitplay;
  play.p := oldp;

  { end of test }
  play.flags  := play.flags and not snd_play;

  data.inchannels  := 0;
  data.outchannels := 0;
  initialized := true;
  chain.next  := nil;
  savemixer;
end;

destructor tsoundblaster.done;
begin
  exitblaster;
  freemem(fpustate,4096);
  restoremixer;
  blaster.available := false;
  initialized := false;
  tobject.done;
end;

function tsoundblaster.getcapabilities(var tcap : tcardinfo) : boolean;
begin
  if not initialized then
  begin
    getcapabilities := false;
    exit;
  end else getcapabilities := true;
  tcap := blaster.cap;
end;

function tsoundblaster.startplayback(freq,flags : longint) : boolean;
begin
  startplayback := false;

  if (flags and snd_auto <> 0) then
  begin
    flags := blaster.cap.flags or flags;
    if (blaster.cap.flags and snd_stereo = 0) then
    flags := flags and not (snd_stereo or snd_prologic);
  end;


  if (flags and snd_16bit <> 0) then flags := flags or snd_signed;
  if (flags and snd_16bit = 0) then flags := flags and not snd_signed;

  if (flags and snd_prologic <> 0) then flags := flags or snd_stereo;
  data.flags := flags;
  case flags and (snd_stereo or snd_prologic) of
    snd_mono   : begin
                   data.inchannels  := 2;
                   data.outchannels := 1;
                 end;
    snd_stereo : begin
                   data.inchannels  := 2;
                   data.outchannels := 2;
                 end;
    snd_stereo+snd_prologic :
                 begin
                   data.inchannels  := 3;
                   data.outchannels := 2;
                 end;
  end;

  if (freq = 0) then { Set frequency automatically }
  begin
    if (flags and snd_stereo <> 0) then freq := blaster.cap.stereoplay.pref
    else freq := blaster.cap.monoplay.pref;
  end;

  if outofspec(freq,flags) then exit;
  data.freq := freq;

  data.samplesize := 1;
  if (flags and snd_16bit <> 0) then inc(data.samplesize,data.samplesize);
  if (flags and snd_stereo <> 0) then inc(data.samplesize,data.samplesize);

  {$IFNDEF FPC}
  play.p := globaldosalloc(data.bufsize*data.samplesize*2);
  getmem32(data.buffer,data.bufsize*data.inchannels*4*2);
  filllong32(ptr48(data.buffer,0),data.bufsize*data.inchannels*2,0);
  {$ELSE}
  play.p := global_dos_alloc(data.bufsize*data.samplesize*2);
  getmem(play.fpcp,data.bufsize*data.samplesize*2);
  getmem(data.buffer,data.bufsize*data.inchannels*4*2);
  filldword(data.buffer^,data.bufsize*data.inchannels*2,0);
  {$ENDIF}

  {$IFNDEF FPC}
  if (data.flags and snd_signed <> 0) then
  begin
    if (data.flags and snd_16bit <> 0) then
      fillword32(ptr48(play.p,0),data.bufsize*data.samplesize,0) else
      fillchar32(ptr48(play.p,0),data.bufsize*data.samplesize*2,0);
  end else
  begin
    if (data.flags and snd_16bit <> 0) then
      fillword32(ptr48(play.p,0),data.bufsize*data.samplesize,32767) else
      fillchar32(ptr48(play.p,0),data.bufsize*data.samplesize*2,127);
  end;
  {$ELSE}
  if (data.flags and snd_signed <> 0) then
  begin
    if (data.flags and snd_16bit <> 0) then
      fillword(play.fpcp^,data.bufsize*data.samplesize,0) else
      fillchar(play.fpcp^,data.bufsize*data.samplesize*2,0);
  end else
  begin
    if (data.flags and snd_16bit <> 0) then
      fillword(play.fpcp^,data.bufsize*data.samplesize,32767) else
      fillchar(play.fpcp^,data.bufsize*data.samplesize*2,127);
  end;
  dpmi_dosmemput(play.p shr 16,0,play.fpcp^,data.bufsize*data.samplesize*2);
  {$ENDIF}
  play.callt := addr(self);
  data.freq := initplay(play.p,data.bufsize*data.samplesize*2,freq,flags);
  startplayback := true;
end;

procedure tsoundblaster.stopplayback;
begin
  if not initialized then exit;
  play.callt := nil;
  exitplay;
  {$IFNDEF FPC}
  globaldosfree(play.p);
  freemem32(data.buffer);
  {$ELSE}
  global_dos_free(play.p);
  freemem(play.fpcp);
  freemem(data.buffer);
  {$ENDIF}
end;

{function tsoundblaster.startrecord(freq,flags : longint) : boolean;
begin
  startrecord := false;

  if (flags and snd_prologic <> 0) then flags := flags or snd_stereo;
  dat.flags := flags;
  case flags and (snd_stereo or snd_prologic) of
    snd_stereo : dat.channels := 2;
    snd_prologic,snd_stereo+snd_prologic : dat.channels := 3;
    else dat.channels := 1;
  end;

  if outofspec(freq,flags) then exit;
  dat.freq := freq;

  dat.samplesize := 1;
  if (flags and snd_16bit <> 0) then inc(dat.samplesize,dat.samplesize);
  if (flags and snd_stereo <> 0) then inc(dat.samplesize,dat.samplesize);

  play.p := globaldosalloc(dat.bufsize*dat.samplesize*2);
  getmem32(dat.buffer,dat.bufsize*dat.channels*4*2);
  filllong32(ptr48(dat.buffer,0),dat.bufsize*dat.channels*2,0);

  if (dat.flags and snd_signed <> 0) then
  begin
    if (dat.flags and snd_16bit <> 0) then
      fillword32(ptr48(play.p,0),dat.bufsize*dat.samplesize,0) else
      fillchar32(ptr48(play.p,0),dat.bufsize*dat.samplesize*2,0);
  end else
  begin
    if (dat.flags and snd_16bit <> 0) then
      fillword32(ptr48(play.p,0),dat.bufsize*dat.samplesize,32767) else
      fillchar32(ptr48(play.p,0),dat.bufsize*dat.samplesize*2,127);
  end;
  play.callt := @self;
  initrecord(rec.p,dat.bufsize*dat.samplesize*2,freq,flags);
  startrecord := true;
end;

procedure tsoundblaster.stoprecord;
begin
  if not initialized then exit;
  rec.callproc := nil;
  exitrecord;
  globaldosfree(rec.p);
  freemem32(dat.buffer);
end;}

procedure tsoundblaster.intrcallproc;
var
  i       : longint;
  ch      : psoundchain;
  p,p2,p3 : pointr;

begin
  {$IFNDEF FPC}
  p  := ptr48(play.p,(data.bufsize*data.samplesize)*play.activeblock);
  p2 := ptr48(data.buffer,(data.bufsize*4*data.inchannels)*play.activeblock);
  {$ELSE}
  p  := pointer(play.fpcp+(data.bufsize*data.samplesize)*play.activeblock);
  p2 := pointer(data.buffer+(data.bufsize*4*data.inchannels)*play.activeblock);
  {$ENDIF}

  { Mix the output channels together }
  mixdata(p2);

  {sat32(p2,data.bufsize*data.inchannels);}
  { Convert mixed data to required output format }
  if (data.flags and snd_prologic <> 0) then
  begin
    {$IFNDEF FPC}
    p3 := ptr48(data.buffer,(data.bufsize*4*data.inchannels)*((play.activeblock+1) and 1));
    {$ELSE}
    p3 := pointer(data.buffer+(data.bufsize*4*data.inchannels)*((play.activeblock+1) and 1));
    {$ENDIF}
    if (data.flags and snd_signed <> 0) then
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_ds16(p2,p3,p,data.bufsize) else
        aud_s32_to_ds8(p2,p3,p,data.bufsize);
    end else
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_dus16(p2,p3,p,data.bufsize) else
        aud_s32_to_dus8(p2,p3,p,data.bufsize);
    end;
  end else
  if (data.flags and snd_stereo <> 0) then
  begin
    if (data.flags and snd_signed <> 0) then
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_s16(p2,p,data.bufsize*data.samplesize shr 1) else
        aud_s32_to_s8(p2,p,data.bufsize*data.samplesize);
    end else
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_us16(p2,p,data.bufsize*data.samplesize shr 1) else
        aud_s32_to_us8(p2,p,data.bufsize*data.samplesize);
    end;
  end else { We play mono :( }
  begin
    if (data.flags and snd_signed <> 0) then
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_sm16(p2,p,data.bufsize) else
        aud_s32_to_sm8(p2,p,data.bufsize);
    end else
    begin
      if (data.flags and snd_16bit <> 0) then
        aud_s32_to_usm16(p2,p,data.bufsize) else
        aud_s32_to_usm8(p2,p,data.bufsize);
    end;
  end;
  {$IFDEF FPC}
  dpmi_dosmemput(play.p shr 16,(data.bufsize*data.samplesize)*play.activeblock,p^,data.bufsize*data.samplesize);
  {$ENDIF}
end;

function tsoundblaster.getversionstring : string;
begin
  getversionstring := versionstring;
end;

function tsoundblaster.getnumofoutcontrols : word;
begin
  getnumofoutcontrols := maxout;
end;

function tsoundblaster.getoutcontroltype(num : word) : word;
begin
  getoutcontroltype := controlmapping[num];
end;

function tsoundblaster.getoutcontrolname(cnt : word) : string;
begin
  if (getnum(cnt) >= maxout) then cnt := cnt_none;
  case cnt of
    cnt_main : getoutcontrolname := 'Master';
    cnt_wave : getoutcontrolname := 'Wave';
    cnt_cd   : getoutcontrolname := 'CD';
    cnt_midi : getoutcontrolname := 'Midi';
    cnt_line : getoutcontrolname := 'Line in';
    cnt_mic  : getoutcontrolname := 'Microphone';
    cnt_spkr : getoutcontrolname := 'PC Speaker';
    cnt_agc  : getoutcontrolname := 'AGC Control';
    cnt_gain : getoutcontrolname := 'Out amplify';
    cnt_3d   : getoutcontrolname := '3D enhancement';

    cnt_trbl : getoutcontrolname := 'Treble';
    cnt_bass : getoutcontrolname := 'Bass';
    else getoutcontrolname := 'N/A';
  end;
end;

function tsoundblaster.getoutcontrolrange(cnt : word) : longint;
begin
  if (cnt = cnt_agc)or(cnt = cnt_3d) then getoutcontrolrange := 1 else
  if (cnt = cnt_gain) then getoutcontrolrange := 3 else
    getoutcontrolrange := maxrange;
end;

function tsoundblaster.getoutcontrol(cnt : word) : longint;
var
  val : word;

procedure getvol(reg : byte);
begin
  if (blaster.cap.version >= sb16) then
  begin
    case (cnt and cnt_chn) of
      cnt_left  : val := readmixer(reg);
      cnt_right : if (reg <> $3A)and(reg <> $3B) then val := readmixer(reg+1);
    end;
    val := val shr 3;
  end else if (blaster.cap.version >= sbpro) then
  begin
    val  := readmixer(reg);
    case (cnt and cnt_chn) of
      cnt_left  : val := val shr 4;
      cnt_right : val := val and $F;
    end;
  end else
  val := readmixer(reg);
{  if (reg = $0A) then left := left shr 1;}
end;

begin
  val := 0;

  if (cnt and cnt_mask = cnt_agc) then
  begin
    if (blaster.cap.version >= sb16) then
    getoutcontrol := readmixer($43) and 1;
    exit;
  end else
  if (cnt and cnt_mask = cnt_agc) then
  begin
    if (blaster.cap.version >= sb16) then
    case (cnt and cnt_chn) of
      cnt_left  : getoutcontrol := readmixer($41) shr 6;
      cnt_right : getoutcontrol := readmixer($42) shr 6;
    end;
    exit;
  end else
  if (cnt and cnt_mask = cnt_3d) then
  begin
    if (blaster.cap.version >= sb16) then
    getoutcontrol := readmixer($90) and 1;
    exit;
  end;

  if (blaster.cap.version >= sb16) then
  begin
    case (cnt and cnt_mask) of
      cnt_main : getvol($30);
      cnt_wave : getvol($32);
      cnt_midi : getvol($34);
      cnt_cd   : getvol($36);
      cnt_line : getvol($38);
      cnt_mic  : getvol($3A);
      cnt_spkr : getvol($3B);
      cnt_trbl : getvol($44);
      cnt_bass : getvol($46);
    end;
  end else if blaster.cap.version >= sbpro then
  begin
    case (cnt and cnt_mask) of
      cnt_main : getvol($22);
      cnt_wave : getvol($04);
      cnt_midi : getvol($26);
      cnt_cd   : getvol($28);
      cnt_line : getvol($2E);
    end;
  end else
  begin
    case (cnt and cnt_mask) of
      cnt_main : getvol($02);
      cnt_wave : getvol($0A);
      cnt_midi : getvol($06);
      cnt_cd   : getvol($08);
    end;
  end;
  getoutcontrol := val;
end;

function tsoundblaster.setoutcontrol(cnt : word;val : longint) : boolean;

  procedure setvol(reg : byte);
  var
    old : byte;

  begin
    if (blaster.cap.version >= sb16) then
    begin
      if (cnt and cnt_right <> 0) then
      begin
        if (reg <> $3A)and(reg <> $3B) then writemixer(reg+1,val);
      end else writemixer(reg,val);
    end else if blaster.cap.version >= sbpro then
    begin
      old := readmixer(reg);
      case (cnt and cnt_chn) of
        cnt_left : val := old and $0F+val shl 4;
        cnt_right : val := old and $F0+val;
      end;
      writemixer(reg,val);
    end else
    begin
      if (reg = $0A)or(reg = $0B) then val := val shr 1;
      writemixer(reg,val);
    end;
  end;

begin
  setoutcontrol := false;

  if (cnt and cnt_mask = cnt_agc) then
  begin
    if (blaster.cap.version < sb16) then exit;
    writemixer($43,readmixer($43) or val);
    setoutcontrol := true;
    exit;
  end else
  if (cnt and cnt_mask = cnt_gain) then
  begin
    if (blaster.cap.version >= sb16) then
    begin
      case (cnt and cnt_chn) of
        cnt_left  : writemixer($41,val shl 6);
        cnt_right : writemixer($42,val shl 6);
      end;
      setoutcontrol := true;
    end;
    exit;
  end else
  if (cnt and cnt_mask = cnt_3d) then
  begin
    if (blaster.cap.version < sb16) then exit;
    writemixer($90,readmixer($90) or val);
    setoutcontrol := true;
    exit;
  end;

  if (val > maxrange) then val := maxrange;

  setoutcontrol := true;

  if (blaster.cap.version >= sb16) then
  begin
    val := val shl 3;
    case (cnt and cnt_mask) of
      cnt_main : setvol($30);
      cnt_wave : setvol($32);
      cnt_midi : setvol($34);
      cnt_cd   : setvol($36);
      cnt_line : setvol($38);
      cnt_mic  : setvol($3A);
      cnt_spkr : setvol($3B);
      cnt_trbl : setvol($44);
      cnt_bass : setvol($46);
      else setoutcontrol := false;
    end;
  end else if (blaster.cap.version >= sbpro) then
  begin
    case (cnt and cnt_mask) of
      cnt_main : setvol($22);
      cnt_wave : setvol($04);
      cnt_midi : setvol($26);
      cnt_cd   : setvol($28);
      cnt_line : setvol($2E);
      else setoutcontrol := false;
    end;
  end else if (blaster.cap.version >= sb21) then
  begin
    case (cnt and cnt_mask) of
      cnt_main : setvol($02);
      cnt_wave : setvol($0A);
      cnt_midi : setvol($06);
      cnt_cd   : setvol($08);
      else setoutcontrol := false;
    end;
  end else setoutcontrol := false;
end;

function tsoundblaster.getnum(cnt : word) : word;
var
  i   : integer;
  num : word;

begin
  num := cnt_none;
  for i := 0 to maxout-1 do if controlmapping[i] = cnt then
  begin
    num := i;
  end;
  getnum := num;
end;

procedure tsoundblaster.savemixer;
var
  i : integer;

begin
  for i := 0 to maxout-1 do
  begin
    settings1[i].l := getoutcontrol(controlmapping[i]+cnt_left);
    settings1[i].r := getoutcontrol(controlmapping[i]+cnt_right);
  end;
  if (blaster.cap.version >= sb16) then
    for i := 1 to 8 do settings2[i] := readmixer($3B+i)
  else if (blaster.cap.version >= sbpro) then
  begin
    settings2[1] := readmixer($0C);
    settings2[2] := readmixer($0E);
  end;
  mixsaved := true;
end;

procedure tsoundblaster.restoremixer;
var
  i : integer;

begin
  if (not mixsaved)or(not autorestore) then exit;
  mixsaved := false;
  for i := 0 to maxout-1 do
  begin
    setoutcontrol(controlmapping[i]+cnt_left,settings1[i].l);
    setoutcontrol(controlmapping[i]+cnt_right,settings1[i].r);
  end;
  if (blaster.cap.version >= sb16) then
  for i := 1 to 8 do writemixer($3B+i,settings2[i])
  else if (blaster.cap.version >= sbpro) then
  begin
    writemixer($0C,settings2[1]);
    writemixer($0E,settings2[2]);
  end;
end;

begin
  blaster.available        := false;
  oldexitproc              := exitproc;
  exitproc                 := addr(exitblaster);
  play.callproc            := dummycaller;
  play.flags               := 0;
  rec.callproc             := dummycaller;
end.
